@@ -1,5 +1,5 @@
 /*
- *  Copyright 2010-2011 Stephen Colebourne
+ *  Copyright 2010-present Stephen Colebourne
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -16,12 +16,20 @@
 package org.joda.convert;
 
 import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertTrue;
+import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.assertSame;
 import static org.junit.Assert.fail;
 
 import java.math.RoundingMode;
 import java.text.ParseException;
 
+import org.joda.convert.test1.Test1Class;
+import org.joda.convert.test2.Test2Class;
+import org.joda.convert.test2.Test2Interface;
+import org.joda.convert.test3.Test3Class;
+import org.joda.convert.test4.Test4Class;
+import org.joda.convert.test4.Test4Interface;
 import org.junit.Test;
 
 /**
@@ -50,6 +58,14 @@ public class TestStringConvert {
         assertEquals(null, conv);
     }
 
+    //-----------------------------------------------------------------------
+    @Test
+    public void test_isConvertible() {
+        assertTrue(StringConvert.INSTANCE.isConvertible(Integer.class));
+        assertTrue(StringConvert.INSTANCE.isConvertible(String.class));
+        assertFalse(StringConvert.INSTANCE.isConvertible(Object.class));
+    }
+
     //-----------------------------------------------------------------------
     @Test
     public void test_convertToString() {
@@ -80,6 +96,13 @@ public class TestStringConvert {
         assertEquals("6", StringConvert.INSTANCE.convertToString(Integer.class, i));
     }
 
+    @Test
+    public void test_convertToString_withType_noGenerics() {
+        Integer i = 6;
+        Class<?> cls = Integer.class;
+        assertEquals("6", StringConvert.INSTANCE.convertToString(cls, i));
+    }
+
     @Test
     public void test_convertToString_withType_primitive1() {
         int i = 6;
@@ -144,6 +167,14 @@ public class TestStringConvert {
     }
 
     //-----------------------------------------------------------------------
+    @Test
+    public void test_findConverter() {
+        Class<Integer> cls = Integer.class;
+        StringConverter<Integer> conv = StringConvert.INSTANCE.findConverter(cls);
+        assertEquals(Integer.valueOf(12), conv.convertFromString(cls, "12"));
+        assertEquals("12", conv.convertToString(12));
+    }
+
     @Test(expected=IllegalArgumentException.class)
     public void test_findConverter_null() {
         StringConvert.INSTANCE.findConverter(null);
@@ -154,6 +185,25 @@ public class TestStringConvert {
         StringConvert.INSTANCE.findConverter(Object.class);
     }
 
+    //-----------------------------------------------------------------------
+    @Test
+    public void test_findConverterNoGenerics() {
+        Class<?> cls = Integer.class;
+        StringConverter<Object> conv = StringConvert.INSTANCE.findConverterNoGenerics(cls);
+        assertEquals(Integer.valueOf(12), conv.convertFromString(cls, "12"));
+        assertEquals("12", conv.convertToString(12));
+    }
+
+    @Test(expected=IllegalArgumentException.class)
+    public void test_findConverterNoGenerics_null() {
+        StringConvert.INSTANCE.findConverterNoGenerics(null);
+    }
+
+    @Test(expected=IllegalStateException.class)
+    public void test_findConverterNoGenerics_Object() {
+        StringConvert.INSTANCE.findConverterNoGenerics(Object.class);
+    }
+
     //-----------------------------------------------------------------------
     @Test
     public void test_convert_annotationMethodMethod() {
@@ -256,7 +306,7 @@ public class TestStringConvert {
         assertEquals("25m", test.convertToString(d));
     }
 
-    // TODO problem is fwks, that just request a converter baed on the type of the object
+    // TODO problem is fwks, that just request a converter based on the type of the object
     @Test(expected = ClassCastException.class)
     public void test_convert_annotationSuperFactorySubViaSub2() {
         StringConvert test = new StringConvert();
@@ -288,6 +338,79 @@ public class TestStringConvert {
         }
     }
 
+    //-----------------------------------------------------------------------
+    @Test
+    public void test_convert_annotationFactoryMethod() {
+        StringConvert test = new StringConvert();
+        DistanceWithFactory d = new DistanceWithFactory(25);
+        assertEquals("25m", test.convertToString(d));
+        assertEquals(d.amount, test.convertFromString(DistanceWithFactory.class, "25m").amount);
+        StringConverter<DistanceWithFactory> conv = test.findConverter(DistanceWithFactory.class);
+        assertEquals(true, conv instanceof MethodsStringConverter<?>);
+        assertSame(conv, test.findConverter(DistanceWithFactory.class));
+        assertEquals(true, conv.toString().startsWith("RefectionStringConverter"));
+    }
+
+    @Test
+    public void test_convert_annotation_ToStringOnInterface() {
+        StringConvert test = new StringConvert();
+        Test1Class d = new Test1Class(25);
+        assertEquals("25g", test.convertToString(d));
+        assertEquals(d.amount, test.convertFromString(Test1Class.class, "25g").amount);
+        StringConverter<Test1Class> conv = test.findConverter(Test1Class.class);
+        assertEquals(true, conv instanceof MethodsStringConverter<?>);
+        assertSame(conv, test.findConverter(Test1Class.class));
+        assertEquals(true, conv.toString().startsWith("RefectionStringConverter"));
+    }
+
+    @Test
+    public void test_convert_annotation_FactoryAndToStringOnInterface() {
+        StringConvert test = new StringConvert();
+        Test2Class d = new Test2Class(25);
+        assertEquals("25g", test.convertToString(d));
+        assertEquals(d.amount, test.convertFromString(Test2Class.class, "25g").amount);
+        StringConverter<Test2Class> conv = test.findConverter(Test2Class.class);
+        assertEquals(true, conv instanceof MethodsStringConverter<?>);
+        assertSame(conv, test.findConverter(Test2Class.class));
+        assertEquals(true, conv.toString().startsWith("RefectionStringConverter"));
+    }
+
+    @Test
+    public void test_convert_annotation_FactoryAndToStringOnInterface_usingInterface() {
+        StringConvert test = new StringConvert();
+        Test2Class d = new Test2Class(25);
+        assertEquals("25g", test.convertToString(d));
+        assertEquals("25g", test.convertFromString(Test2Interface.class, "25g").print());
+        StringConverter<Test2Interface> conv = test.findConverter(Test2Interface.class);
+        assertEquals(true, conv instanceof MethodsStringConverter<?>);
+        assertSame(conv, test.findConverter(Test2Interface.class));
+        assertEquals(true, conv.toString().startsWith("RefectionStringConverter"));
+    }
+
+    @Test
+    public void test_convert_annotation_ToStringFromStringOnSuperClassBeatsInterface() {
+        StringConvert test = new StringConvert();
+        Test3Class d = new Test3Class(25);
+        assertEquals("25g", test.convertToString(d));
+        assertEquals(d.amount, test.convertFromString(Test3Class.class, "25g").amount);
+        StringConverter<Test3Class> conv = test.findConverter(Test3Class.class);
+        assertEquals(true, conv instanceof MethodsStringConverter<?>);
+        assertSame(conv, test.findConverter(Test3Class.class));
+        assertEquals(true, conv.toString().startsWith("RefectionStringConverter"));
+    }
+
+    @Test(expected=IllegalStateException.class)
+    public void test_convert_annotation_FromStringFactoryClashingMethods_fromClass() {
+        StringConvert test = new StringConvert();
+        test.findConverter(Test4Class.class);
+    }
+
+    @Test(expected=IllegalStateException.class)
+    public void test_convert_annotation_FromStringFactoryClashingMethods_fromInterface() {
+        StringConvert test = new StringConvert();
+        test.findConverter(Test4Interface.class);
+    }
+
     //-----------------------------------------------------------------------
     @Test(expected=IllegalStateException.class)
     public void test_convert_annotationNoMethods() {
@@ -383,7 +506,7 @@ public class TestStringConvert {
         StringConvert.INSTANCE.register(Integer.class, MockIntegerStringConverter.INSTANCE);
     }
 
-    @Test(expected=IllegalStateException.class)
+    @Test
     public void test_register_classAlreadyRegistered() {
         new StringConvert().register(Integer.class, MockIntegerStringConverter.INSTANCE);
     }
@@ -394,6 +517,48 @@ public class TestStringConvert {
         assertSame(MockDistanceStringConverter.INSTANCE, test.findConverter(DistanceMethodMethod.class));
     }
 
+    //-------------------------------------------------------------------------
+    ToStringConverter<DistanceNoAnnotations> DISTANCE_TO_STRING_CONVERTER = new ToStringConverter<DistanceNoAnnotations>() {
+        public String convertToString(DistanceNoAnnotations object) {
+            return object.toString();
+        }
+    };
+    FromStringConverter<DistanceNoAnnotations> DISTANCE_FROM_STRING_CONVERTER = new FromStringConverter<DistanceNoAnnotations>() {
+        public DistanceNoAnnotations convertFromString(Class<? extends DistanceNoAnnotations> cls, String str) {
+            return DistanceNoAnnotations.parse(str);
+        }
+    };
+
+    @Test
+    public void test_register_FunctionalInterfaces() {
+        StringConvert test = new StringConvert();
+        test.register(DistanceNoAnnotations.class, DISTANCE_TO_STRING_CONVERTER, DISTANCE_FROM_STRING_CONVERTER);
+        DistanceNoAnnotations d = new DistanceNoAnnotations(25);
+        assertEquals("Distance[25m]", test.convertToString(d));
+        assertEquals(d.amount, test.convertFromString(DistanceNoAnnotations.class, "25m").amount);
+        StringConverter<DistanceNoAnnotations> conv = test.findConverter(DistanceNoAnnotations.class);
+        assertEquals(true, conv.getClass().getName().contains("$"));
+        assertSame(conv, test.findConverter(DistanceNoAnnotations.class));
+    }
+
+    @Test(expected=IllegalArgumentException.class)
+    public void test_register_FunctionalInterfaces_nullClass() {
+        StringConvert test = new StringConvert();
+        test.register(null, DISTANCE_TO_STRING_CONVERTER, DISTANCE_FROM_STRING_CONVERTER);
+    }
+
+    @Test(expected=IllegalArgumentException.class)
+    public void test_register_FunctionalInterfaces_nullToString() {
+        StringConvert test = new StringConvert();
+        test.register(DistanceNoAnnotations.class, null, DISTANCE_FROM_STRING_CONVERTER);
+    }
+
+    @Test(expected=IllegalArgumentException.class)
+    public void test_register_FunctionalInterfaces_nullFromString() {
+        StringConvert test = new StringConvert();
+        test.register(DistanceNoAnnotations.class, DISTANCE_TO_STRING_CONVERTER, null);
+    }
+
     //-------------------------------------------------------------------------
     @Test
     public void test_registerMethods() {
@@ -437,7 +602,7 @@ public class TestStringConvert {
         test.registerMethods(DistanceNoAnnotations.class, "toString", null);
     }
 
-    @Test(expected=IllegalStateException.class)
+    @Test
     public void test_registerMethods_classAlreadyRegistered() {
       StringConvert test = new StringConvert();
       test.registerMethods(DistanceNoAnnotations.class, "toString", "parse");
@@ -481,7 +646,7 @@ public class TestStringConvert {
         test.registerMethodConstructor(DistanceNoAnnotations.class, null);
     }
 
-    @Test(expected=IllegalStateException.class)
+    @Test
     public void test_registerMethodConstructor_classAlreadyRegistered() {
       StringConvert test = new StringConvert();
       test.registerMethodConstructor(DistanceNoAnnotations.class, "toString");
